#include <SPI.h>
#include <TFT.h>
#include <Keypad.h>

const int SW_pin = 7;
const int X_pin = 0;
const int Y_pin = 1;

const byte ROWS = 4;
const byte COLS = 4;

char keys[ROWS][COLS] = {
  {'1', '2', '3', 'A'},
  {'4', '5', '6', 'B'},
  {'7', '8', '9', 'C'},
  {'*', '0', '#', 'D'}
};

#define CS 10
#define DC 9
#define RST 12
#define SUDOKU_SIZE 9
#define NUM_ELEM (SUDOKU_SIZE*SUDOKU_SIZE)

byte rowPins[ROWS] = {A5, A4, A3, A2};
byte colPins[COLS] = {5, 4, 3, 2};

Keypad keypad = Keypad(makeKeymap(keys), rowPins, colPins, ROWS, COLS);

int start_boards[6][NUM_ELEM] = {
  {0, 0, 5, 0, 6, 0, 4, 0, 0,
   9, 7, 0, 4, 0, 3, 6, 0, 0,
   0, 0, 4, 0, 9, 0, 2, 0, 3,
   0, 1, 8, 5, 3, 0, 0, 6, 0,
   0, 0, 0, 1, 2, 9, 0, 0, 0,
   0, 2, 0, 0, 8, 7, 5, 3, 0,
   8, 0, 3, 0, 7, 0, 1, 0, 0,
   0, 0, 7, 8, 0, 5, 0, 9, 4,
   0, 0, 1, 0, 4, 0, 7, 0, 0},
  {0, 2, 9, 0, 0, 0, 0, 6, 4,
   3, 0, 6, 0, 4, 1, 8, 0, 0,
   0, 0, 0, 0, 2, 9, 0, 1, 3,
   0, 1, 8, 3, 0, 0, 4, 0, 2,
   0, 0, 3, 1, 0, 4, 6, 0, 0,
   7, 0, 4, 0, 0, 2, 1, 3, 0,
   0, 3, 0, 2, 7, 0, 0, 0, 0,
   0, 0, 2, 8, 1, 0, 3, 0, 6,
   6, 5, 0, 0, 0, 0, 2, 8, 0},
  {5, 6, 3, 7, 0, 4, 9, 0, 0,
   0, 8, 0, 0, 0, 0, 0, 2, 7,
   0, 0, 7, 0, 0, 6, 3, 0, 0,
   0, 0, 8, 6, 0, 0, 4, 9, 2,
   0, 4, 0, 0, 3, 0, 0, 6, 0,
   6, 5, 9, 0, 0, 1, 7, 0, 0,
   0, 0, 6, 4, 0, 0, 5, 0, 0,
   9, 1, 0, 0, 0, 0, 0, 7, 0,
   0, 0, 5, 9, 0, 3, 2, 4, 6},
  {0, 0, 0, 6, 0, 0, 0, 0, 8,
   0, 0, 4, 0, 0, 7, 0, 5, 0,
   0, 6, 0, 8, 0, 0, 0, 0, 0,
   0, 0, 7, 0, 0, 0, 3, 0, 5,
   5, 0, 0, 2, 9, 3, 0, 0, 7,
   4, 0, 2, 0, 0, 0, 9, 0, 0,
   0, 0, 0, 0, 0, 5, 0, 4, 0,
   0, 2, 0, 4, 0, 0, 7, 0, 0,
   8, 0, 0, 0, 0, 9, 0, 0, 0},
  {7, 0, 0, 0, 6, 2, 0, 0, 0,
   0, 5, 0, 0, 8, 0, 0, 0, 0,
   0, 9, 0, 0, 0, 4, 1, 0, 0,
   8, 2, 0, 0, 0, 0, 3, 0, 0,
   0, 0, 5, 0, 9, 0, 4, 0, 0,
   0, 0, 3, 0, 0, 0, 0, 5, 8,
   0, 0, 1, 8, 0, 0, 0, 2, 0,
   0, 0, 0, 0, 3, 0, 0, 8, 0,
   0, 0, 0, 2, 5, 0, 0, 0, 3},
  {1, 0, 5, 0, 6, 8, 4, 7, 0,
   9, 7, 2, 4, 5, 3, 6, 1, 8,
   6, 8, 4, 7, 9, 1, 2, 5, 3,
   7, 1, 8, 5, 3, 4, 9, 6, 2,
   3, 5, 6, 1, 2, 9, 8, 4, 7,
   4, 2, 9, 6, 8, 7, 5, 3, 1,
   8, 4, 3, 9, 7, 6, 1, 2, 5,
   2, 6, 7, 8, 1, 5, 3, 9, 4,
   5, 9, 1, 3, 4, 2, 7, 8, 6}
};

TFT TFTscreen = TFT(CS, DC, RST);

int cell_height;
int cell_width;
int pos_x;
int pos_y;
int cells_filled;
int starting = 1;


int sudoku_board[NUM_ELEM];
byte blocked[NUM_ELEM];

void draw_sudoku() {
  TFTscreen.stroke(255, 255, 255);
  TFTscreen.setTextSize(1);
  int i, j;
  char text[2];
  //draw cells
  for (i = 0; i < SUDOKU_SIZE; i++) {
    for (j = 0; j < SUDOKU_SIZE; j++) {
      if (i != pos_x || j != pos_y){
        TFTscreen.rect(j * cell_width, i * cell_height, cell_width, cell_height);
        if (sudoku_board[i * SUDOKU_SIZE + j] > 0 && blocked[i * SUDOKU_SIZE + j]) {
          TFTscreen.stroke(255, 255, 255);
          sprintf(text, "%d\0", sudoku_board[i * SUDOKU_SIZE + j]);
          TFTscreen.text(text, j * cell_width + cell_width / 3, i * cell_height + cell_height / 4);
        } else {
          TFTscreen.fill(50, 0, 50);
          TFTscreen.rect(j * cell_width, i * cell_height, cell_width, cell_height);
          TFTscreen.noFill();
          if (sudoku_board[i * SUDOKU_SIZE + j] != 0) {
            TFTscreen.stroke(255, 255, 255);
            sprintf(text, "%d\0", sudoku_board[i * SUDOKU_SIZE + j]);
            TFTscreen.text(text, j * cell_width + cell_width / 3, i * cell_height + cell_height / 4);
          }
        }
      }
    }
  }
  
  TFTscreen.stroke(255, 0, 0);
  TFTscreen.rect(pos_y * cell_width, pos_x * cell_height, cell_width, cell_height);
  if (sudoku_board[pos_x * SUDOKU_SIZE + pos_y] > 0) {
    sprintf(text, "%d\0", sudoku_board[pos_x * SUDOKU_SIZE + pos_y]);
    TFTscreen.text(text, pos_y * cell_width + cell_width / 3, pos_x * cell_height + cell_height / 4);
  }

}

void draw_grid() {
  TFTscreen.stroke(255, 0, 255);
  TFTscreen.rect(0 * cell_width, 0 * cell_height, 3 * cell_width, 3 * cell_height);
  TFTscreen.rect(0 * cell_width, 3 * cell_height, 3 * cell_width, 3 * cell_height);
  TFTscreen.rect(0 * cell_width, 6 * cell_height, 3 * cell_width, 3 * cell_height);
  TFTscreen.rect(3 * cell_width, 0 * cell_height, 3 * cell_width, 3 * cell_height);
  TFTscreen.rect(3 * cell_width, 3 * cell_height, 3 * cell_width, 3 * cell_height);
  TFTscreen.rect(3 * cell_width, 6 * cell_height, 3 * cell_width, 3 * cell_height);
  TFTscreen.rect(6 * cell_width, 0 * cell_height, 3 * cell_width, 3 * cell_height);
  TFTscreen.rect(6 * cell_width, 3 * cell_height, 3 * cell_width, 3 * cell_height);
  TFTscreen.rect(6 * cell_width, 6 * cell_height, 3 * cell_width, 3 * cell_height);
}

void set_super_easy_board() {
  cells_filled = 0;
  int i, j;
  int* board = start_boards[random(5, 5)];
  for (i = 0; i < SUDOKU_SIZE; i++) {
    for (j = 0; j < SUDOKU_SIZE; j++) {
      sudoku_board[i * SUDOKU_SIZE + j] = board[i * SUDOKU_SIZE + j];
      if (sudoku_board[i * SUDOKU_SIZE + j] != 0) {
        cells_filled ++;
        blocked[i * SUDOKU_SIZE + j] = 1;
      } else {
        blocked[i * SUDOKU_SIZE + j] = 0;
      }
    }
  }
  
  draw_sudoku();
  draw_grid();
}

void set_easy_board() {
  cells_filled = 0;
  int i, j;
  int* board = start_boards[random(0, 2)];
  for (i = 0; i < SUDOKU_SIZE; i++) {
    for (j = 0; j < SUDOKU_SIZE; j++) {
      sudoku_board[i * SUDOKU_SIZE + j] = board[i * SUDOKU_SIZE + j];
      if (sudoku_board[i * SUDOKU_SIZE + j] != 0) {
        cells_filled ++;
        blocked[i * SUDOKU_SIZE + j] = 1;
      } else {
        blocked[i * SUDOKU_SIZE + j] = 0;
      }
    }
  }
  
  draw_sudoku();
  draw_grid();
}

void set_hard_board() {
  cells_filled = 0;
  int i, j;
  int* board = start_boards[random(3, 4)];
  for (i = 0; i < SUDOKU_SIZE; i++) {
    for (j = 0; j < SUDOKU_SIZE; j++) {
      sudoku_board[i * SUDOKU_SIZE + j] = board[i * SUDOKU_SIZE + j];
      if (sudoku_board[i * SUDOKU_SIZE + j] != 0) {
        cells_filled ++;
        blocked[i * SUDOKU_SIZE + j] = 1;
      } else {
        blocked[i * SUDOKU_SIZE + j] = 0;
      }
    }
  }
  
  draw_sudoku();
  draw_grid();
}

void write_key(char key) {
  if (!blocked[pos_x * SUDOKU_SIZE + pos_y]) {
    cells_filled ++;
    sudoku_board[pos_x * SUDOKU_SIZE + pos_y] = key - '0';
    TFTscreen.fill(125, 50, 50);
    TFTscreen.stroke(255, 255, 255);
    TFTscreen.rect(pos_y * cell_width, pos_x * cell_height, cell_width, cell_height);
    TFTscreen.noFill();
  }
}

void check_win() {
  int win = 1;
  if (cells_filled < NUM_ELEM) {
    TFTscreen.background(0, 0, 0);
    TFTscreen.setTextSize(2);
    TFTscreen.stroke(255, 255, 255);
    TFTscreen.text("NOT FINISHED!", 5, TFTscreen.height() / 2 - 10);
    delay(3000);
    TFTscreen.background(0, 0, 0);
    draw_sudoku();
    draw_grid();
  } else {
    int i,j,k;
    byte visited[SUDOKU_SIZE + 1];

    // correct rows?
    for (i = 0; i < SUDOKU_SIZE; i++) {
      for (k = 1; k <= 9; k++)
      visited[k] = 0;
      for (j = 0; j < SUDOKU_SIZE; j++) {
        visited[sudoku_board[i * SUDOKU_SIZE + j]] ++;
      }
      for (k = 1; k <= 9; k++) {
        if (visited[k] == 0)
          win = 0;
      }
    }

    // correct columns?
    for (i = 0; i < SUDOKU_SIZE; i++) {
      for (k = 1; k <= 9; k++)
      visited[k] = 0;
      for (j = 0; j < SUDOKU_SIZE; j++) {
        visited[sudoku_board[j * SUDOKU_SIZE + i]] ++;
      }
      for (k = 1; k <= 9; k++) {
        if (visited[k] == 0)
          win = 0;
      }
    }

    // correct 3x3 squares?
    int m, n;
    for (i = 0; i < 3; i++) {
      for (j = 0; j < 3; j++){
        for (k = 1; k <= 9; k++)
        visited[k] = 0;
        for (m = 0; m < 3; m++) {
          for (n = 0; n < 3; n++){
            visited[sudoku_board[(i * 3 + m) * SUDOKU_SIZE + (j * 3 + n)]] ++;
          }
        }
        for (k = 1; k <= 9; k++) {
          if (visited[k] == 0)
            win = 0;
        }
      }
    }

    if (win == 1) {
      TFTscreen.background(0, 0, 0);
      TFTscreen.setTextSize(2);
      TFTscreen.stroke(255, 255, 255);
      TFTscreen.text("YOU WON!", TFTscreen.width() / 2 - 50, TFTscreen.height() / 2 - 20);
      TFTscreen.text("CONGRATS!", TFTscreen.width() / 2 - 55, TFTscreen.height() / 2 + 10);
      delay(5000);
      draw_menu();
    } else {
      TFTscreen.background(0, 0, 0);
      TFTscreen.setTextSize(2);
      TFTscreen.stroke(255, 255, 255);
      TFTscreen.text("TRY AGAIN!", 5, TFTscreen.height() / 2 - 10);
      delay(3000);
      TFTscreen.background(0, 0, 0);
      draw_sudoku();
      draw_grid();
    }
  }
}

void draw_menu() {
  starting = 1;
  TFTscreen.background(0, 0, 0);
  TFTscreen.setTextSize(4);
  TFTscreen.stroke(0, 255, 255);
  TFTscreen.text("S", 10, 10);
  TFTscreen.stroke(255, 0, 0);
  TFTscreen.text("U", 33, 10);
  TFTscreen.stroke(0, 255, 0);
  TFTscreen.text("D", 56, 10);
  TFTscreen.stroke(0, 0, 255);
  TFTscreen.text("O", 79, 10);
  TFTscreen.stroke(255, 255, 0);
  TFTscreen.text("K", 102, 10);
  TFTscreen.stroke(255, 0, 255);
  TFTscreen.text("U", 125, 10);
  TFTscreen.setTextSize(1);
  TFTscreen.stroke(255, 255, 255);
  TFTscreen.text("Press 'A' for      GAME", 10, TFTscreen.height() / 2 - 10);
  TFTscreen.text("Press 'B' for      GAME", 10, TFTscreen.height() / 2 + 10);
  TFTscreen.stroke(0, 255, 0);
  TFTscreen.text("EASY", 94, TFTscreen.height() / 2 - 10);
  TFTscreen.stroke(255, 0, 0);
  TFTscreen.text("HARD", 94, TFTscreen.height() / 2 + 10);
  TFTscreen.stroke(255, 255, 255);
  TFTscreen.setTextSize(0.5);
  TFTscreen.text("Click '*' to check your", 10, TFTscreen.height() / 2 + 40);
  TFTscreen.text("game!", TFTscreen.width() / 2 - 15, TFTscreen.height() / 2 + 50);
}

void setup() {
  pinMode(SW_pin, INPUT);
  digitalWrite(SW_pin, HIGH);
  Serial.begin(9600);

  TFTscreen.begin();
  TFTscreen.background(0, 0, 0);
  TFTscreen.setRotation(3);

  cell_height = TFTscreen.height() / SUDOKU_SIZE;
  cell_width = TFTscreen.width() / SUDOKU_SIZE;
  pos_x = 0;
  pos_y = 0;

  draw_menu();
}

void loop() {
  // put your main code here, to run repeatedly:
  char key = keypad.getKey();
  if (starting == 0) {
    if (analogRead(X_pin) == 1023) {
      pos_y = (pos_y + 1) % SUDOKU_SIZE;
      draw_sudoku();
      draw_grid();
    }
    if (analogRead(X_pin) == 0) {
      if (pos_y >= 1) {
        pos_y -= 1;
        draw_sudoku();
        draw_grid();
      }
    }
    if (analogRead(Y_pin) == 1023) {
      pos_x = (pos_x + 1) % SUDOKU_SIZE;
      draw_sudoku();
      draw_grid();
    }
    if (analogRead(Y_pin) == 0) {
      if (pos_x >= 1) {
        pos_x -= 1;
        draw_sudoku();
        draw_grid();
      }
    }
    delay(100);
  }

  if (starting == 1) {
    if (key) {
      if (key == 'A') {
        TFTscreen.background(0, 0, 0);
        set_easy_board();
      }
      if (key == 'B') {
        TFTscreen.background(0, 0, 0);
        set_hard_board();
      }
      if (key == 'C') {
        TFTscreen.background(0, 0, 0);
        set_super_easy_board();
      }
      starting = 0;
    }
  } else {
    if (key) {
      Serial.println(key);
      if (key >= '1' && key <= '9') {
        write_key(key);
      }
      if (key == '*') {
        check_win();
      }
    }
  }
}
